"""
Tests for instance entity from the API.

Copyright 2017 ICTU

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import unittest
from mock import MagicMock
from bigboat.client import Client
from bigboat.application import Application
from bigboat.instance import Instance

class ApplicationTest(unittest.TestCase):
    """
    Tests for the application instance entity.
    """

    def setUp(self):
        self.client = MagicMock(spec_set=Client)
        self.application = Application(self.client, 'nginx', 'latest')
        self.instance = Instance(self.client, 'nginx', current_state='starting',
                                 desired_state='running',
                                 application=self.application,
                                 services={'www': {'state': 'starting'}},
                                 parameters={'SETTING': 'value'},
                                 options={'storageBucket': 'custom'})

    def test_update(self):
        """
        Test the Instance.update method.
        """

        # Test that instances without application information cannot be started
        incomplete_instance = Instance(self.client, 'foo')
        with self.assertRaises(ValueError):
            incomplete_instance.update()

        value = self.instance.update()
        update_instance = self.client.update_instance
        update_instance.assert_called_once_with('nginx', 'nginx', 'latest',
                                                parameters={
                                                    'SETTING': 'value'
                                                },
                                                options={
                                                    'storageBucket': 'custom'
                                                })
        self.assertEqual(value, update_instance.return_value)

    def test_delete(self):
        """
        Test the Instance.delete method.
        """

        value = self.instance.delete()
        self.client.delete_instance.assert_called_once_with('nginx')
        self.assertEqual(value, self.client.delete_instance.return_value)

    def test_repr(self):
        """
        Test the Instance.__repr__ method.
        """

        application = "Application(name='nginx', version='latest')"
        services = "{'www': {'state': 'starting'}}"
        instance = "Instance(name='nginx', current_state='starting', " + \
                   "desired_state='running', application=" + \
                   application + ", services=" + services + ", " + \
                   "parameters={'SETTING': 'value'}, " + \
                   "options={'storageBucket': 'custom'})"
        self.assertEqual(repr(self.instance), instance)



"""
Tests for utilities.

Copyright 2017 ICTU

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import unittest
from bigboat.utils import readonly, Inherited as inherit

@readonly(['name', 'version'], rest='other')
class Item(object):
    """
    Item with some readonly attributes.
    """

    def __init__(self):
        self._name = 'foo'
        self._version = '1'
        self._rest = 'data'

    def get(self):
        """
        Retrieve data.
        """

        return self._name

    @classmethod
    def execute(cls):
        """
        Perform some dummy task.

        Returns:
            bool: Success value
        """

        return True

    def __getattr__(self, name):
        raise AttributeError

class Subitem(Item):
    """
    Extending class.
    """

    @inherit
    def get(self):
        """
        Inheriting method which has custom documentation.
        """

        return self._version

    @inherit
    def get_other(self):
        """
        New method that is incorrrecty marked as inheriting.
        """

        return self._rest

    @inherit
    @classmethod
    def execute(cls):
        return False

    @inherit
    @classmethod
    def new_execute(cls): # pylint: disable=missing-docstring
        # New classmethod that is incorrectly marked as inheriting, but it has
        # no documentation block itself.
        return None

class Utils_Test(unittest.TestCase):
    """
    Tests for utilities.
    """

    def test_readonly(self):
        """
        Test the readonly decorator.
        """

        item = Item()
        self.assertEqual(item.name, 'foo')
        self.assertEqual(item.version, '1')
        self.assertEqual(item.other, 'data')
        with self.assertRaises(AttributeError):
            dummy = item.nonexistent

    def test_inherit(self):
        """
        Test the Inherited decorator.
        """

        item = Item()
        sub_item = Subitem()
        # Custom documentation on inheriting method is retained.
        self.assertEqual(sub_item.get.__doc__.strip(),
                         'Inheriting method which has custom documentation.')
        self.assertEqual(item.get(), 'foo')
        self.assertEqual(sub_item.get(), '1')
        self.assertEqual(sub_item.get_other(), 'data')

        self.assertEqual(Subitem.execute.__doc__, Item.execute.__doc__)
        self.assertTrue(Item.execute())
        self.assertFalse(Subitem.execute())
        self.assertIsNone(Subitem.new_execute())



"""
Clients that connect to the BigBoat API.

Copyright 2017 ICTU

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from builtins import str
from builtins import object
import requests
import yaml
from .application import Application
from .instance import Instance
from .utils import Inherited as inherit

class Client(object):
    """
    Generic client base class, enforcing minimum required interface.
    """

    def __init__(self, base_url, **kwargs):
        self._base_url = base_url.rstrip('/')
        self._options = kwargs

    @property
    def base_url(self):
        """
        The base URL of the BigBoat instance.
        """

        return self._base_url

    def apps(self):
        """
        Retrieve all application definitions from the API.

        Returns:
            :obj:`list` of :obj:`application.Application`
        """

        raise NotImplementedError('Must be implemented by subclasses')

    def get_app(self, name, version):
        """
        Retrieve a specific application definition from the API.

        Args:
            name (str): The name of the application
            version (str): The version of the application

        Returns:
            :obj:`bigboat.application.Application` or `None`: The application
            definition if it was found or `None` if the definition does not
            exist.
        """

        raise NotImplementedError('Must be implemented by subclasses')

    def update_app(self, name, version):
        """
        Register an application definition in the API.

        Args:
            name (str): The name of the application
            version (str): The version of the application

        Returns:
            :obj:`bigboat.application.Application` or `None`: The application
            definition if it was successfully created.
        """

        raise NotImplementedError('Must be implemented by subclasses')

    def delete_app(self, name, version):
        """
        Delete an application definition in the API.

        Args:
            name (str): The name of the application
            version (str): The version of the application

        Returns:
            bool: Whether the application was successfully deleted.
        """

        raise NotImplementedError('Must be implemented by subclasses')

    def instances(self):
        """
        Retrieve all live instances from the API.

        Returns:
            :obj:`list` of :obj:`bigboat.instance.Instance`
        """

        raise NotImplementedError('Must be implemented by subclasses')

    def get_instance(self, name):
        """
        Retrieve a specific live instance from the API.

        Args:
            name (str): The name of the instance.

        Returns:
            :obj:`bigboat.instance.Instance` or `None`: The instance
            if it was found or `None` if the instance does not exist.
        """

        raise NotImplementedError('Must be implemented by subclasses')

    def update_instance(self, name, app_name, version, **kwargs):
        """
        Request the instance to be created with a desired state of 'running'.

        Args:
            name (str): The name of the instance to be started.
            app_name (str): The name of the application to be started.
            version (str): The version of the application to be started.
            **kwargs: Additional properties to use when starting the instance.

        Returns:
            :obj:`bigboat.instance.Instance` or `None`: The instance
            if it was started or `None` if the instance failed to start.
        """

        raise NotImplementedError('Must be implemented by subclasses')

    def delete_instance(self, name):
        """
        Retrieve a specific live instance from the API.

        Args:
            name (str): The name of the instance.

        Returns:
            :obj:`bigboat.instance.Instance` or `None`: The instance
            if it was found or `None` if the instance does not exist.
        """

        raise NotImplementedError('Must be implemented by subclasses')

class Client_v1(Client):
    """
    Client for the deprecated BigBoat v1 API.
    """

    def _format_url(self, path):
        return '{}/api/v1/{}'.format(self._base_url, path)

    def _get(self, path):
        return requests.get(self._format_url(path))

    def _delete(self, path):
        return requests.delete(self._format_url(path))

    @inherit
    def apps(self):
        return []

    @inherit
    def get_app(self, name, version):
        try:
            request = self._get('appdef/{}/{}'.format(name, version))
        except requests.exceptions.ConnectionError:
            return None

        document = yaml.load(request.text)

        return Application(self, document['name'], str(document['version']))

    @inherit
    def update_app(self, name, version):
        # Cannot create new apps through v1 API
        return None

    @inherit
    def delete_app(self, name, version):
        request = self._delete('appdef/{}/{}'.format(name, version))

        if request.status_code == 404:
            return False

        return request.status_code == 200

    @inherit
    def instances(self):
        request = self._get('instances')

        if request.status_code == 404:
            return []

        data = request.json()
        return [Instance(self, name) for name in data['instances']]

    @inherit
    def get_instance(self, name):
        request = self._get('state/{}'.format(name))

        if request.status_code == 404:
            return None

        state = 'running' if request.text == 'active' else request.text

        return Instance(self, name, state)

    @inherit
    def update_instance(self, name, app_name, version, **kwargs):
        request = self._get('start-app/{}/{}/{}'.format(app_name, version, name))

        if request.status_code == 404:
            return None

        return Instance(self, name, current_state='running',
                        application=Application(self, app_name, version))

    @inherit
    def delete_instance(self, name):
        request = self._get('stop-app/{}'.format(name))

        if request.status_code == 404:
            return None

        return Instance(self, name, 'created')

class Client_v2(Client):
    """
    Client for the BigBoat v2 API.
    """

    def __init__(self, base_url, api_key):
        super(Client_v2, self).__init__(base_url)
        self._api_key = api_key
        self._session = requests.Session()
        self._session.headers.update({'api-key': self._api_key})

    def _format_url(self, path):
        return '{}/api/v2/{}'.format(self._base_url, path)

    def _get(self, path):
        return self._session.get(self._format_url(path))

    def _put(self, path, content_type=None, data=None, json=None):
        headers = {}
        if content_type is not None:
            headers['Content-Type'] = content_type
        elif json is not None:
            headers['Content-Type'] = 'application/json'

        return self._session.put(self._format_url(path), headers=headers,
                                 data=data, json=json)

    def _delete(self, path):
        return self._session.delete(self._format_url(path))

    @staticmethod
    def _check_bad_request(request):
        # Bad Request should raise an exception
        if request.status_code == 400:
            if request.headers['Content-Type'] == 'application/json':
                response = request.json()
                raise ValueError(response['message'])
            else:
                raise ValueError(request.text)

    def _format_app(self, app):
        return Application(self, app['name'], app['version'])

    @inherit
    def apps(self):
        request = self._get('apps')
        return [self._format_app(app) for app in request.json()]

    @inherit
    def get_app(self, name, version):
        request = self._get('apps/{}/{}'.format(name, version))
        if request.status_code == 404:
            return None

        return self._format_app(request.json())

    @inherit
    def update_app(self, name, version):
        try:
            request = self._put('apps/{}/{}'.format(name, version))
        except requests.exceptions.ConnectionError:
            return None

        return self._format_app(request.json())

    @inherit
    def delete_app(self, name, version):
        request = self._delete('apps/{}/{}'.format(name, version))
        if request.status_code == 404:
            return False

        return True

    def get_compose(self, name, version, file_name):
        """
        Retrieve a docker compose or bigboat compose file for the application.

        Args:
            name (str): The name of the application
            version (str): The version of the application
            file_name (str): 'dockerCompose' or 'bigboatCompose'

        Returns:
            :obj:`str` or `None`: The application definition's docker compose
            file contents if the application was found, or `None` if the
            definition does not exist.
        """

        path = 'apps/{}/{}/files/{}'.format(name, version, file_name)
        request = self._get(path)
        if request.status_code == 404:
            return None

        content_type = request.headers.get('content-type')
        if content_type not in ('text/plain', 'text/yaml'):
            return None

        return request.text

    def update_compose(self, name, version, file_name, content):
        """
        Update a docker compose or bigboat compose file for the application.

        Args:
            name (str): The name of the application
            version (str): The version of the application
            file_name (str): 'dockerCompose' or 'bigboatCompose'
            content (str): The file contents

        Returns:
            bool: Whether the compose file was successfully updated.

        Raises:
            ValueError: When the compose file could not be parsed as a valid
            YAML file.
            ValueError: When the bigboatCompose file contains name or version
            properties that do not match the provided application name/verison.
        """

        path = 'apps/{}/{}/files/{}'.format(name, version, file_name)
        request = self._put(path, content_type='text/plain', data=content)
        if request.status_code == 404:
            return False

        self._check_bad_request(request)

        if request.status_code != 201:
            return False

        return True

    def _format_instance(self, instance):
        if 'app' in instance:
            application = self._format_app(instance['app'])
        else:
            application = None

        services = instance.get('services')

        state = instance.get('state', {})

        return Instance(self, instance.get('name'),
                        current_state=state.get('current', 'running'),
                        desired_state=state.get('desired'),
                        application=application, services=services)

    @inherit
    def instances(self):
        request = self._get('instances')
        return [self._format_instance(instance) for instance in request.json()]

    @inherit
    def get_instance(self, name):
        request = self._get('instances/{}'.format(name))

        if request.status_code == 404:
            return None

        return self._format_instance(request.json())

    @inherit
    def update_instance(self, name, app_name, version, **kwargs):
        data = {
            'app': app_name,
            'version': version,
            'parameters': kwargs.get('parameters') or {},
            'options': kwargs.get('options') or {}
        }
        request = self._put('instances/{}'.format(name), json=data)

        self._check_bad_request(request)

        return self._format_instance(request.json())

    @inherit
    def delete_instance(self, name):
        request = self._delete('instances/{}'.format(name))

        self._check_bad_request(request)

        return self._format_instance(request.json())

    def statuses(self):
        """
        Retrieve all status items reported by BigBoat.

        Returns:
            :obj:`list` of :obj:`dict`: The status items
        """

        return self._get('status').json()



"""
Tests for clients that connect to the BigBoat API.

Copyright 2017 ICTU

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import json
import unittest
import requests
import requests_mock
import yaml
from bigboat.client import Client, Client_v1, Client_v2

class Client_Test(unittest.TestCase):
    """
    Tests for the generic client base class.
    """

    URL = 'http://dashboard.example/'
    CANONICAL_URL = 'http://dashboard.example'

    def setUp(self):
        super(Client_Test, self).setUp()
        self.client = Client(self.URL, other='data')

    def test_base_url(self):
        """
        Test the Client.base_url property.
        """

        # The base_url should return an URL without trailing slash.
        self.assertEqual(self.client.base_url, self.CANONICAL_URL)

    def test_interface(self):
        """
        Test whether the abstract base class requires the minimal interface.
        """

        with self.assertRaises(NotImplementedError):
            self.client.apps()
        with self.assertRaises(NotImplementedError):
            self.client.get_app('foo', 'latest')
        with self.assertRaises(NotImplementedError):
            self.client.update_app('foo', 'latest')
        with self.assertRaises(NotImplementedError):
            self.client.delete_app('foo', 'latest')
        with self.assertRaises(NotImplementedError):
            self.client.instances()
        with self.assertRaises(NotImplementedError):
            self.client.get_instance('bar')
        with self.assertRaises(NotImplementedError):
            self.client.update_instance('bar', 'foo', 'latest', extra='yes')
        with self.assertRaises(NotImplementedError):
            self.client.delete_instance('bar')

class RequestsTestCase(unittest.TestCase):
    """
    A base unit test class that uses requests.

    This class provides a requests_mock member variable during the test run.
    """

    def setUp(self):
        super(RequestsTestCase, self).setUp()
        self.requests_mock = requests_mock.mock(case_sensitive=True)
        self.requests_mock.start()

    def tearDown(self):
        super(RequestsTestCase, self).tearDown()
        self.requests_mock.stop()

class Client_v1_Test(RequestsTestCase):
    """
    Tests for the BigBoat v1 API.
    """

    URL = 'http://dashboard.example/'
    PATH = 'api/v1/'

    def setUp(self):
        super(Client_v1_Test, self).setUp()
        self.client = Client_v1(self.URL)

    def test_apps(self):
        """
        Test the Client_v1.apps dummy method.
        """

        self.assertEqual(self.client.apps(), [])

    def test_get_app(self):
        """
        Test the Client_v1.get_app method.
        """

        self.requests_mock.get(self.URL + self.PATH + 'appdef/foo/latest',
                               text='''name: foo
version: latest
instance:
  image: hello-world
''')
        application = self.client.get_app('foo', 'latest')
        self.assertEqual(application.name, 'foo')
        self.assertEqual(application.version, 'latest')

        # Test connection error for nonexistent applications.
        self.requests_mock.get(self.URL + self.PATH + 'appdef/does/notexist',
                               exc=requests.exceptions.ConnectionError)
        self.assertIsNone(self.client.get_app('does', 'notexist'))

    def test_update_app(self):
        """
        Test the Client_v1.update_app dummy method.
        """

        self.assertIsNone(self.client.update_app('foo', 'latest'))

    def test_delete_app(self):
        """
        Test the Client_v1.delete_app method.
        """

        self.requests_mock.delete(self.URL + self.PATH + 'appdef/foo/latest')
        self.requests_mock.delete(self.URL + self.PATH + 'appdef/does/notexist',
                                  status_code=404)

        self.assertTrue(self.client.delete_app('foo', 'latest'))
        self.assertFalse(self.client.delete_app('does', 'notexist'))

    def test_instances(self):
        """
        Test the Client_v1.instances method.
        """

        self.requests_mock.get(self.URL + self.PATH + 'instances',
                               status_code=404)
        self.assertEqual(self.client.instances(), [])

        self.requests_mock.get(self.URL + self.PATH + 'instances',
                               json={
                                   "statusCode": 200,
                                   "instances": ['foo', 'bar', 'baz']
                               })
        instances = self.client.instances()
        names = list(sorted(instance.name for instance in instances))
        self.assertEqual(names, ['bar', 'baz', 'foo'])

    def test_get_instance(self):
        """
        Test the Client_v1.get_instance method.
        """

        self.requests_mock.get(self.URL + self.PATH + 'state/qux',
                               status_code=404)
        self.requests_mock.get(self.URL + self.PATH + 'state/foo',
                               text='active')

        self.assertIsNone(self.client.get_instance('qux'))
        instance = self.client.get_instance('foo')
        self.assertEqual(instance.name, 'foo')
        self.assertEqual(instance.current_state, 'running')

    def test_update_instance(self):
        """
        Test the update_instance method.
        """

        url = self.URL + self.PATH
        self.requests_mock.get(url + 'start-app/does/notexist/qux',
                               status_code=404)
        self.requests_mock.get(url + 'start-app/foo/latest/foo')

        self.assertIsNone(self.client.update_instance('qux', 'does', 'notexist',
                                                      more='args'))
        instance = self.client.update_instance('foo', 'foo', 'latest')
        self.assertEqual(instance.name, 'foo')
        self.assertEqual(instance.current_state, 'running')
        self.assertEqual(instance.application.name, 'foo')
        self.assertEqual(instance.application.version, 'latest')

    def test_delete_instance(self):
        """
        Test the Client_v1.delete_instance method.
        """

        self.requests_mock.get(self.URL + self.PATH + 'stop-app/qux',
                               status_code=404)
        self.requests_mock.get(self.URL + self.PATH + 'stop-app/foo')

        self.assertIsNone(self.client.delete_instance('qux'))
        instance = self.client.delete_instance('foo')
        self.assertEqual(instance.name, 'foo')
        self.assertEqual(instance.current_state, 'created')

class Client_v2_Test(RequestsTestCase):
    """
    Tests for the BigBoat v2 API.
    """

    URL = 'http://dashboard.example/'
    PATH = 'api/v2/'
    KEY = 'my-api-key'

    def setUp(self):
        super(Client_v2_Test, self).setUp()
        self.client = Client_v2(self.URL, self.KEY)

    def test_apps(self):
        """
        Test the Client_v2.apps method.
        """

        self.requests_mock.get(self.URL + self.PATH + 'apps',
                               json=[
                                   {
                                       "id": "ERfrBncoPKSN9ampt",
                                       "name": "nginx",
                                       "version": "latest"
                                   },
                                   {
                                       "id": "LENn6QcjnG8WRvAxf",
                                       "name": "nginx",
                                       "version": "1.11.4"
                                   }
                               ])

        apps = self.client.apps()
        app_pairs = list(sorted((app.name, app.version) for app in apps))
        self.assertEqual(app_pairs, [('nginx', '1.11.4'), ('nginx', 'latest')])

    def test_get_app(self):
        """
        Test the Client_v2.get_app method.
        """

        self.requests_mock.get(self.URL + self.PATH + 'apps/does/notexist',
                               status_code=404)
        self.requests_mock.get(self.URL + self.PATH + 'apps/nginx/latest',
                               json={
                                   "id": "MKMZCnLcEJmkjSenJ",
                                   "name": "nginx",
                                   "version": "latest"
                               })

        self.assertIsNone(self.client.get_app('does', 'notexist'))

        app = self.client.get_app('nginx', 'latest')
        self.assertEqual(app.name, 'nginx')
        self.assertEqual(app.version, 'latest')

    def test_update_app(self):
        """
        Test the Client_v2.update_app method.
        """

        self.requests_mock.put(self.URL + self.PATH + 'apps/%3Cfoo%3E/0',
                               exc=requests.exceptions.ConnectionError)
        self.requests_mock.put(self.URL + self.PATH + 'apps/nginx/latest',
                               status_code=201,
                               json={
                                   "id": "MKMZCnLcEJmkjSenJ",
                                   "name": "nginx",
                                   "version": "latest"
                               })

        self.assertIsNone(self.client.update_app('<foo>', '0'))
        app = self.client.update_app('nginx', 'latest')
        self.assertEqual(app.name, 'nginx')
        self.assertEqual(app.version, 'latest')

    def test_delete_app(self):
        """
        Test the Client_v2.delete_app method.
        """

        self.requests_mock.delete(self.URL + self.PATH + 'apps/does/notexist',
                                  status_code=404)
        self.requests_mock.delete(self.URL + self.PATH + 'apps/nginx/latest',
                                  status_code=204)

        self.assertFalse(self.client.delete_app('does', 'notexist'))
        self.assertTrue(self.client.delete_app('nginx', 'latest'))

    def test_get_compose(self):
        """
        Test the Client_v2.get_compose method.
        """

        url = self.URL + self.PATH
        content = '''name: nginx
version: latest

www:
  enable_ssh: true'''
        self.requests_mock.get(url + 'apps/does/notexist/files/dockerCompose',
                               status_code=404)
        self.requests_mock.get(url + 'apps/nginx/latest/files/notUsed',
                               headers={'Content-Type': 'text/html'},
                               text='<html><body>No such path</body></html')
        self.requests_mock.get(url + 'apps/nginx/latest/files/bigboatCompose',
                               headers={'Content-Type': 'text/plain'},
                               text=content)

        self.assertIsNone(self.client.get_compose('does', 'notexist',
                                                  'dockerCompose'))
        self.assertIsNone(self.client.get_compose('nginx', 'latest', 'notUsed'))
        self.assertEqual(self.client.get_compose('nginx', 'latest',
                                                 'bigboatCompose'), content)

    @staticmethod
    def _put_bigboat_compose_handler(request, context):
        yaml_file = request.body
        try:
            data = yaml.load(yaml_file)
        except yaml.error.YAMLError as yaml_error:
            context.status_code = 400
            context.headers = 'text/plain'
            return "Problem asserting validity of YAML: {}".format(yaml_error)

        if data['name'] != 'nginx':
            context.status_code = 400
            context.headers['content-type'] = 'application/json'
            return json.dumps({
                'message': 'Name property of Bigboat compose needs to be equal '
                           'to name property of App'
            })

        context.status_code = 201
        return yaml_file

    def test_update_compose(self):
        """
        Test the Client_v2.update_compose method.
        """

        url = self.URL + self.PATH
        self.requests_mock.put(url + 'apps/does/notexist/files/dockerCompose',
                               status_code=404)
        self.requests_mock.put(url + 'apps/nginx/latest/files/notUsed',
                               headers={'Content-Type': 'text/html'},
                               text='<html><body>No such path</body></html')
        self.requests_mock.put(url + 'apps/nginx/latest/files/bigboatCompose',
                               text=self._put_bigboat_compose_handler)

        self.assertFalse(self.client.update_compose('does', 'notexist',
                                                    'dockerCompose', 'x: y'))
        self.assertFalse(self.client.update_compose('nginx', 'latest',
                                                    'notUsed', 'x: y'))
        with self.assertRaises(ValueError):
            self.client.update_compose('nginx', 'latest', 'bigboatCompose', ':')
        with self.assertRaises(ValueError):
            self.client.update_compose('nginx', 'latest', 'bigboatCompose',
                                       'name: somethingElse\nversion: latest')

        content = '''name: nginx
version: latest

www:
  enable_ssh: true'''
        self.assertTrue(self.client.update_compose('nginx', 'latest',
                                                   'bigboatCompose', content))

    def test_instances(self):
        """
        Test the Client_v2.instances method.
        """

        self.requests_mock.get(self.URL + self.PATH + 'instances',
                               json=[
                                   {
                                       "id": "y7bzwghzP9ouM56g6",
                                       "name": "nginx",
                                       "state": {
                                           "current": "running",
                                           "desired": "running"
                                       }
                                   },
                                   {
                                       "id": "ySYXNTPw2ry9XE6nu",
                                       "name": "nginx2",
                                       "state": {
                                           "current": "starting",
                                           "desired": "running"
                                       }
                                   }
                               ])

        instances = self.client.instances()
        instance_data = list(sorted([
            (instance.name, instance.current_state, instance.desired_state)
            for instance in instances
        ]))
        self.assertEqual(instance_data, [
            ('nginx', 'running', 'running'),
            ('nginx2', 'starting', 'running')
        ])

    def test_get_instance(self):
        """
        Test the Client_v2.get_instance method.
        """

        self.requests_mock.get(self.URL + self.PATH + 'instances/qux',
                               status_code=404)
        self.requests_mock.get(self.URL + self.PATH + 'instances/nginx',
                               json={
                                   "id": "y7bzwghzP9ouM56g6",
                                   "name": "nginx",
                                   "state": {
                                       "current": "starting",
                                       "desired": "running"
                                   },
                                   "app": {
                                       "name": "nginx",
                                       "version": "latest"
                                   },
                                   "services": {
                                       "www": {
                                           "state": "starting"
                                       }
                                   }
                               })

        self.assertIsNone(self.client.get_instance('qux'))
        instance = self.client.get_instance('nginx')
        self.assertEqual(instance.name, 'nginx')
        self.assertEqual(instance.current_state, 'starting')
        self.assertEqual(instance.desired_state, 'running')
        self.assertEqual(instance.application.name, 'nginx')
        self.assertEqual(instance.application.version, 'latest')
        self.assertEqual(instance.services, {'www': {'state': 'starting'}})

    def test_update_instance(self):
        """
        Test the Client_v2.update_instance method.
        """

        self.requests_mock.put(self.URL + self.PATH + 'instances/error',
                               status_code=400, text='error',
                               headers={'content-type': 'text/plain'})
        self.requests_mock.put(self.URL + self.PATH + 'instances/nginx',
                               json={
                                   "id": "y7bzwghzP9ouM56g6",
                                   "name": "nginx",
                                   "state": {
                                       "current": "starting",
                                       "desired": "running"
                                   },
                                   "app": {
                                       "name": "nginx",
                                       "version": "latest"
                                   },
                                   "services": {
                                       "www": {
                                           "state": "starting"
                                       }
                                   }
                               })

        with self.assertRaises(ValueError):
            self.client.update_instance('error', 'does', 'notexist')

        instance = self.client.update_instance('nginx', 'nginx', 'latest',
                                               parameters={
                                                   "SETTING": "value"
                                               },
                                               options={
                                                   "storageBucket": "custom"
                                               })

        self.assertEqual(instance.name, 'nginx')
        self.assertEqual(instance.application.name, 'nginx')
        self.assertEqual(instance.application.version, 'latest')
        self.assertEqual(instance.current_state, 'starting')
        self.assertEqual(instance.desired_state, 'running')
        self.assertEqual(instance.services, {'www': {'state': 'starting'}})

    def test_delete_instance(self):
        """
        Test the Client_v2.delete_instance method.
        """

        self.requests_mock.delete(self.URL + self.PATH + 'instances/error',
                                  status_code=400, text='error',
                                  headers={'content-type': 'text/plain'})
        self.requests_mock.delete(self.URL + self.PATH + 'instances/nginx',
                                  json={
                                      "id": "y7bzwghzP9ouM56g6",
                                      "name": "nginx",
                                      "state": {
                                          "current": "stopping",
                                          "desired": "stopped"
                                      },
                                      "app": {
                                          "name": "nginx",
                                          "version": "latest"
                                      },
                                      "services": {
                                          "www": {
                                              "state": "stopping"
                                          }
                                      }
                                  })

        with self.assertRaises(ValueError):
            self.client.delete_instance('error')

        instance = self.client.delete_instance('nginx')
        self.assertEqual(instance.name, 'nginx')
        self.assertEqual(instance.application.name, 'nginx')
        self.assertEqual(instance.application.version, 'latest')
        self.assertEqual(instance.current_state, 'stopping')
        self.assertEqual(instance.desired_state, 'stopped')
        self.assertEqual(instance.services, {'www': {'state': 'stopping'}})

    def test_statuses(self):
        """
        Test the Client_v2.statuses method.
        """

        content = [
            {
                "name": "Available IPs",
                "lastCheck": {
                    "time": 1494245442228,
                    "ISO": "2017-05-08T12:10:42.228Z"
                },
                "description": "Total number of IPs: 190. IPs in use: 19. "
                               "<strong>Available IPs: 171</strong>",
                "details": {
                    "totalIps": 190,
                    "usedIps": 19
                },
                "isOk": True
            },
            {
                "name": "Docker graph: /var/lib/docker",
                "lastCheck": {
                    "time": 1494245487722,
                    "ISO": "2017-05-08T12:11:27.722Z"
                },
                "description": "Total size: 218.2 GB. "
                               "<strong>Available: 96.3 GB</strong>",
                "details": {
                    "total": 234322399232,
                    "used": 130956759040,
                    "free": 103365640192
                },
                "isOk": True
            },
        ]

        self.requests_mock.get(self.URL + self.PATH + 'status',
                               json=content)

        self.assertEqual(self.client.statuses(), content)



"""
Instance entity from the API.

Copyright 2017 ICTU

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from .entity import Entity
from .utils import readonly

@readonly("name", "current_state", "desired_state", "application", "services",
          "parameters", "options")
class Instance(Entity):
    """
    A deployed (parameterized) application instance entity.
    """

    def __init__(self, client, name, current_state=None, **kwargs):
        super(Instance, self).__init__(client)
        self._name = name
        self._current_state = current_state
        self._desired_state = kwargs.get('desired_state')
        self._application = kwargs.get('application')
        self._services = kwargs.get('services')

        self._parameters = kwargs.get('parameters')
        self._options = kwargs.get('options')

    def update(self):
        """
        Request the instance to be created with a desired state of 'running'.
        """

        if self.application is None:
            raise ValueError('Application information required to start instance')

        return self.client.update_instance(self.name, self.application.name,
                                           self.application.version,
                                           parameters=self.parameters,
                                           options=self.options)

    def delete(self):
        """
        Request the instance to be stopped.
        """

        return self.client.delete_instance(self.name)

    def __repr__(self):
        parts = [
            ('name', self.name),
            ('current_state', self.current_state),
            ('desired_state', self.desired_state),
            ('application', self.application),
            ('services', self.services),
            ('parameters', self.parameters),
            ('options', self.options)
        ]
        properties = ['{}={!r}'.format(key, value) for (key, value) in parts]

        return 'Instance({})'.format(', '.join(properties))



"""
Tests for generic base entity.

Copyright 2017 ICTU

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import unittest
from mock import MagicMock
from bigboat.client import Client
from bigboat.entity import Entity

class Entity_Test(unittest.TestCase):
    """
    Tests for entity from the BigBoat API.
    """

    def test_interface(self):
        """
        Test the enforcement of abstract interface.
        """

        client = MagicMock(spec_set=Client)
        entity = Entity(client)

        # Test the client property.
        self.assertEqual(entity.client, client)

        # Test the interface
        with self.assertRaises(NotImplementedError):
            entity.update()

        with self.assertRaises(NotImplementedError):
            entity.delete()

        # Test unknown attribute access
        with self.assertRaises(AttributeError):
            dummy = entity.nonexistent



#!/usr/bin/env python

"""
Package setup script.

Copyright 2017 ICTU

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from builtins import str
from pip.download import PipSession
from pip.req import parse_requirements
from setuptools import setup, find_packages
from bigboat import __version__

setup(name='bigboat',
      version=__version__,
      description='BigBoat docker dashboard API',
      long_description='''Python wrapper library for the BigBoat API.
Support for v2 and the deprecated v1 is included.
This API can create, retrieve, update and delete application definitions,
do similar operations for instances and poll for status''',
      author='ICTU',
      author_email='leon.helwerda@ictu.nl',
      url='https://github.com/ICTU/bigboat-python-api',
      license='Apache License, Version 2.0',
      packages=find_packages(),
      scripts=[],
      include_package_data=True,
      install_requires=[
          str(requirement.req)
          for requirement in parse_requirements('requirements.txt',
                                                session=PipSession())
      ],
      test_suite='tests',
      classifiers=[
          'Development Status :: 3 - Alpha',
          'Environment :: Web Environment',
          'Intended Audience :: Developers',
          'License :: OSI Approved :: Apache Software License',
          'Operating System :: OS Independent',
          'Programming Language :: Python',
          'Programming Language :: Python :: 2.7',
          'Programming Language :: Python :: 3.6',
          'Topic :: Internet :: WWW/HTTP',
          'Topic :: Software Development :: Libraries :: Python Modules',
          'Topic :: Software Development :: Build Tools'],
      keywords=['docker', 'dashboard', 'bigboat', 'api'])



"""
Generic base enntity.

Copyright 2017 ICTU

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from builtins import object
from .utils import readonly

@readonly("client")
class Entity(object):
    """
    An entity from the BigBoat API.
    """

    def __init__(self, client):
        self._client = client

    def update(self):
        """
        Send the data of the entity as an update for its properties to the
        BigBoat API. The entity calls the relevant update method of the client.

        Returns:
            :obj:`bigboat.entity.Entity` or `None`: The updated entity object
            from the API response if it was successfully updated.
        """

        raise NotImplementedError('Must be implemented by subclasses')

    def delete(self):
        """
        Delete the entity from the BigBoat API. The entity calls the relevant
        update method of the client.

        Returns:
            bool: Whether the entity was successfully deleted.
        """

        raise NotImplementedError('Must be implemented by subclasses')

    def __getattr__(self, attr):
        raise AttributeError



"""
Utilities in the BigBoat API library.

Copyright 2017 ICTU

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from functools import partial, wraps, WRAPPER_ASSIGNMENTS
from past.builtins import basestring

def readonly(*args, **kwargs):
    """
    Register readonly properties for member variables of a class instance.

    Args:
        *args: Variable length list of properties to register as providers of
            read-only access to protected member variables with the same name,
            prefixed with an underscore '_'.
        **kwargs: Arbitrary keyword arguments of properties (keywords) to
            register as providers of read-only access to protected member
            variables with the same name as the argument value.

    Returns:
        A decorator function that applies on a class to provide read-only
        member properties.
    """

    if args and not isinstance(args[0], basestring):
        properties = args[0]
    else:
        properties = args

    aliased_properties = kwargs

    def decorator(subject):
        """
        Register the properties in the class.

        Args:
            subject: The class instance.

        Returns:
            The altered class instance.
        """

        def _get_property(property_name, instance):
            return getattr(instance, property_name)

        for property_name in properties:
            setattr(subject, property_name,
                    property(fget=partial(_get_property, '_' + property_name)))
        for variable_name, property_name in aliased_properties.items():
            setattr(subject, property_name,
                    property(fget=partial(_get_property, '_' + variable_name)))

        return subject

    return decorator

class Inherited(object):
    """
    Indicate that an inherited method whose parent method has documentation.

    This solution is based on the 'Docstring inheritance decorator' Python
    recipe from http://code.activestate.com/recipes/576862/ which is licensed
    under the MIT License (but none of the code was used verbatim).
    """

    _wrappers = tuple(prop for prop in WRAPPER_ASSIGNMENTS if prop != '__doc__')

    def __init__(self, method):
        self._method = method

    def __get__(self, im_self, im_class):
        if im_self is not None:
            method, overridden, doc = self.get_object_parent(im_class, im_self)
        else:
            method, overridden, doc = self.get_class_parent(im_class)

        # Prefer original documentation from the inheriting method.
        if doc is not None:
            method.__doc__ = doc
        elif overridden is not None:
            method.__doc__ = overridden.__doc__

        return method

    def get_object_parent(self, im_class, im_self):
        """
        Determine the parent method using the class and instance context.

        Args:
            im_class: The method's class instance.
            im_self: The method's bound object instance.

        Returns:
            A tuple:
            - The wrapped method.
            - The parent method or `None` if it was not found.
            - The original documentation of the method or `None` if there is no
              documentation on the inheriting method.
        """

        @wraps(self._method, assigned=self._wrappers)
        def wrapper(*args, **kwargs):
            """
            Wrapper for the actual method.
            """

            return self._method(im_self, *args, **kwargs)

        parent = super(im_class, im_self)
        overridden = getattr(parent, self._method.__name__, None)
        return wrapper, overridden, self._method.__doc__

    def get_class_parent(self, im_class):
        """
        Determine the parent method using the class context.

        Args:
            im_class: The method's class instance.

        Returns:
            A tuple:
            - The wrapped method.
            - The parent method or `None` if it was not found.
            - The original documentation of the method or `None` if there is no
              documentation on the inheriting method.
        """

        func = self._method.__func__
        name = func.__name__
        doc = func.__doc__

        @wraps(func, assigned=self._wrappers)
        def wrapper(*args, **kwargs):
            """
            Wrapper for the actual function behind the classmehod.
            """

            return func(im_class, *args, **kwargs)

        parents = im_class.__mro__[1:]
        for parent in parents:
            overridden = getattr(parent, name, None)
            if overridden is not None:
                return wrapper, overridden, doc

        return wrapper, None, doc



"""
Application entity from the API.

Copyright 2017 ICTU

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from .entity import Entity
from .utils import readonly

@readonly("name", "version")
class Application(Entity):
    """
    An application definition entity.
    """

    def __init__(self, client, name, version):
        super(Application, self).__init__(client)

        self._name = name
        self._version = version

    def update(self):
        """
        Register the application definition in the BigBoat API.
        """

        return self.client.update_app(self.name, self.version)

    def delete(self):
        """
        Delete the application definition in the BigBoat API.
        """

        return self.client.delete_app(self.name, self.version)

    def start(self, name=None, **kwargs):
        """
        Request an instance to be created with a desired state of 'running'
        for this application.

        Args:
            name (:obj:`str` or `None`): The name of the instance to be
                started, or `None` to use the application name.
            **kwargs: Additional properties to use when starting the instance.

        Returns:
            :obj:`bigboat.instance.Instance` or `None`: The instance
            if it was started or `None` if the instance failed to start.
        """

        if name is None:
            name = self.name

        return self.client.update_instance(name, self.name, self.version,
                                           **kwargs)

    def __repr__(self):
        return 'Application(name={!r}, version={!r})'.format(self.name, self.version)
